"use strict";

var _graphql = require("../graphql");

var _ = require("..");

var _graphqlVersion = require("../utils/graphqlVersion");

var _type = require("../type");

var _ObjectTypeComposer = require("../ObjectTypeComposer");

var _InputTypeComposer = require("../InputTypeComposer");

var _ScalarTypeComposer = require("../ScalarTypeComposer");

var _EnumTypeComposer = require("../EnumTypeComposer");

var _InterfaceTypeComposer = require("../InterfaceTypeComposer");

var _UnionTypeComposer = require("../UnionTypeComposer");

var _Resolver = require("../Resolver");

var _TypeMapper = require("../TypeMapper");

var _ListComposer = require("../ListComposer");

var _NonNullComposer = require("../NonNullComposer");

var _ThunkComposer = require("../ThunkComposer");

let typeMapper;
beforeEach(() => {
  _.schemaComposer.clear();

  typeMapper = new _TypeMapper.TypeMapper(_.schemaComposer);
});
describe('TypeMapper', () => {
  it('should add basic scalar GraphQL types', () => {
    expect(typeMapper.schemaComposer.get('String').getType()).toBe(_graphql.GraphQLString);
    expect(typeMapper.schemaComposer.get('Float').getType()).toBe(_graphql.GraphQLFloat);
    expect(typeMapper.schemaComposer.get('Int').getType()).toBe(_graphql.GraphQLInt);
    expect(typeMapper.schemaComposer.get('Boolean').getType()).toBe(_graphql.GraphQLBoolean);
    expect(typeMapper.schemaComposer.get('ID').getType()).toBe(_graphql.GraphQLID);
    expect(typeMapper.schemaComposer.get('JSON').getType()).toBe(_type.GraphQLJSON);
    expect(typeMapper.schemaComposer.get('JSONObject').getType()).toBe(_type.GraphQLJSONObject);
    expect(typeMapper.schemaComposer.get('Date').getType()).toBe(_type.GraphQLDate);
    expect(typeMapper.schemaComposer.get('Buffer').getType()).toBe(_type.GraphQLBuffer);
  });
  it('should allow to override basic graphql-compose types', () => {
    const CustomJSON = _.schemaComposer.createScalarTC({
      name: 'CustomJSON',
      serialize: () => {}
    });

    const CustomDate = _.schemaComposer.createScalarTC({
      name: 'CustomDate',
      serialize: () => {}
    });

    const CustomBuffer = _.schemaComposer.createScalarTC({
      name: 'CustomBuffer',
      serialize: () => {}
    });

    _.schemaComposer.set('JSON', CustomJSON);

    _.schemaComposer.set('Date', CustomDate);

    _.schemaComposer.set('Buffer', CustomBuffer);

    expect(typeMapper.schemaComposer.get('JSON')).toBe(CustomJSON);
    expect(typeMapper.schemaComposer.get('Date')).toBe(CustomDate);
    expect(typeMapper.schemaComposer.get('Buffer')).toBe(CustomBuffer);
  });
  it('should create object type from template string', () => {
    const tc = typeMapper.convertSDLTypeDefinition(_graphqlVersion.graphqlVersion < 12 ? `
          type IntRange {
            # Max value
            max: Int,
            min: Int!
            arr: [String]
          }
        ` : `
          type IntRange {
            """Max value"""
            max: Int,
            min: Int!
            arr: [String]
          }
        `);
    expect(tc).toBeInstanceOf(_ObjectTypeComposer.ObjectTypeComposer);
    const type = tc.getType();
    expect(type).toBeInstanceOf(_graphql.GraphQLObjectType);
    expect(typeMapper.schemaComposer.get('IntRange')).toBe(tc);
    const IntRangeTC = new _ObjectTypeComposer.ObjectTypeComposer(type, _.schemaComposer);
    expect(IntRangeTC.getTypeName()).toBe('IntRange');
    expect(IntRangeTC.getFieldNames()).toEqual(expect.arrayContaining(['max', 'min', 'arr']));
    expect(IntRangeTC.getFieldType('max')).toBe(_graphql.GraphQLInt);
    expect(IntRangeTC.getFieldConfig('max').description).toBe('Max value');
    expect(IntRangeTC.getFieldType('min')).toBeInstanceOf(_graphql.GraphQLNonNull);
    expect(IntRangeTC.getFieldType('arr')).toBeInstanceOf(_graphql.GraphQLList);
  });
  it('should create input object type from template string', () => {
    const tc = typeMapper.convertSDLTypeDefinition(`
      input InputIntRange {
        min: Int @default(value: 0)
        max: Int!
      }
    `);
    expect(tc).toBeInstanceOf(_InputTypeComposer.InputTypeComposer);
    const type = tc.getType();
    expect(type).toBeInstanceOf(_graphql.GraphQLInputObjectType);
    expect(typeMapper.schemaComposer.get('InputIntRange')).toBe(tc);
    const IntRangeTC = new _InputTypeComposer.InputTypeComposer(type, _.schemaComposer);
    expect(IntRangeTC.getTypeName()).toBe('InputIntRange');
    expect(IntRangeTC.getField('min').defaultValue).toBe(0);
    expect(IntRangeTC.getFieldType('min')).toBe(_graphql.GraphQLInt);
    expect(IntRangeTC.getFieldType('max')).toBeInstanceOf(_graphql.GraphQLNonNull);
    expect(IntRangeTC.getFieldType('max').ofType).toBe(_graphql.GraphQLInt);
  });
  it('should create interface type from template string', () => {
    const tc = typeMapper.convertSDLTypeDefinition(_graphqlVersion.graphqlVersion < 12 ? `
          interface IntRangeInterface {
            # Max value
            max: Int,
            min: Int!
            arr: [String]
          }
        ` : `
          interface IntRangeInterface {
            """Max value"""
            max: Int,
            min: Int!
            arr: [String]
          }
        `);
    expect(tc).toBeInstanceOf(_InterfaceTypeComposer.InterfaceTypeComposer);
    const type = tc.getType();
    expect(type).toBeInstanceOf(_graphql.GraphQLInterfaceType);
    expect(typeMapper.schemaComposer.get('IntRangeInterface').getType()).toBe(type);
    const IntRangeTC = new _InterfaceTypeComposer.InterfaceTypeComposer(type, _.schemaComposer);
    expect(IntRangeTC.getTypeName()).toBe('IntRangeInterface');
    expect(IntRangeTC.getFieldNames()).toEqual(expect.arrayContaining(['max', 'min', 'arr']));
    expect(IntRangeTC.getFieldType('max')).toBe(_graphql.GraphQLInt);
    expect(IntRangeTC.getFieldConfig('max').description).toBe('Max value');
    expect(IntRangeTC.getFieldType('min')).toBeInstanceOf(_graphql.GraphQLNonNull);
    expect(IntRangeTC.getFieldType('arr')).toBeInstanceOf(_graphql.GraphQLList);
  });
  it('should return wrapped type', () => {
    expect(typeMapper.convertSDLWrappedTypeName('String!')).toBeInstanceOf(_NonNullComposer.NonNullComposer);
    expect(typeMapper.convertSDLWrappedTypeName('[String]')).toBeInstanceOf(_ListComposer.ListComposer);
    expect(typeMapper.convertSDLWrappedTypeName('[String]!')).toBeInstanceOf(_NonNullComposer.NonNullComposer);
    expect(typeMapper.convertSDLWrappedTypeName('[String]!').ofType).toBeInstanceOf(_ListComposer.ListComposer);
    expect(typeMapper.convertSDLWrappedTypeName('String')).toBeInstanceOf(_ScalarTypeComposer.ScalarTypeComposer);
    expect(typeMapper.convertSDLWrappedTypeName('String').getType()).toBe(_graphql.GraphQLString);
  });
  describe('convertOutputFieldConfig()', () => {
    describe('converting field type', () => {
      it('should accept type with GraphQLOutputType', () => {
        const fc = typeMapper.convertOutputFieldConfig({
          type: _graphql.GraphQLString
        });
        expect(fc.type).toBeInstanceOf(_ScalarTypeComposer.ScalarTypeComposer);
        expect(fc.type.getType()).toBe(_graphql.GraphQLString);
        const objectType = new _graphql.GraphQLObjectType({
          name: 'SomeType',
          fields: {
            f: {
              type: _graphql.GraphQLString
            }
          }
        });
        const fc2 = typeMapper.convertOutputFieldConfig({
          type: objectType
        });
        expect(fc2.type).toBeInstanceOf(_ObjectTypeComposer.ObjectTypeComposer);
        expect(fc2.type.getType()).toBe(objectType);
      });
      it('should accept GraphQLScalarType', () => {
        const fc = typeMapper.convertOutputFieldConfig(_graphql.GraphQLString);
        expect(fc.type).toBeInstanceOf(_ScalarTypeComposer.ScalarTypeComposer);
        expect(fc.type.getType()).toBe(_graphql.GraphQLString);
      });
      it('should accept GraphQLObjectType', () => {
        const type = new _graphql.GraphQLObjectType({
          name: 'Test',
          fields: () => ({
            a: {
              type: _graphql.GraphQLInt
            }
          })
        });
        const fc = typeMapper.convertOutputFieldConfig(type);
        expect(fc.type).toBeInstanceOf(_ObjectTypeComposer.ObjectTypeComposer);
        expect(fc.type.getType()).toBe(type);
      });
      it('should accept GraphQLNonNull', () => {
        const fc = typeMapper.convertOutputFieldConfig(new _graphql.GraphQLNonNull(_graphql.GraphQLString));
        expect(fc.type).toBeInstanceOf(_NonNullComposer.NonNullComposer);
        expect(fc.type.getType().ofType).toBe(_graphql.GraphQLString);
      });
      it('should accept GraphQLList', () => {
        const fc = typeMapper.convertOutputFieldConfig(new _graphql.GraphQLList(_graphql.GraphQLString));
        expect(fc.type).toBeInstanceOf(_ListComposer.ListComposer);
        expect(fc.type.getType().ofType).toBe(_graphql.GraphQLString);
      });
      it('should accept type as string to scalar', () => {
        const fc = typeMapper.convertOutputFieldConfig({
          type: 'String'
        });
        expect(fc.type).toBeInstanceOf(_ScalarTypeComposer.ScalarTypeComposer);
        expect(fc.type.getType()).toBe(_graphql.GraphQLString);
      });
      it('should accept type as string to object type', () => {
        _ObjectTypeComposer.ObjectTypeComposer.create('type Foo { id: ID! }', _.schemaComposer);

        const fc = typeMapper.convertOutputFieldConfig({
          type: 'Foo'
        });
        expect(fc.type).toBe(_.schemaComposer.getAnyTC('Foo'));
      });
      it('should accept type as string to enum type', () => {
        _EnumTypeComposer.EnumTypeComposer.create('enum Foo { A B }', _.schemaComposer);

        const fc = typeMapper.convertOutputFieldConfig({
          type: 'Foo'
        });
        expect(fc.type).toBe(_.schemaComposer.getAnyTC('Foo'));
      });
      it('should accept type as string to interface type', () => {
        _InterfaceTypeComposer.InterfaceTypeComposer.create('interface Foo { id: ID! }', _.schemaComposer);

        const fc = typeMapper.convertOutputFieldConfig({
          type: 'Foo'
        });
        expect(fc.type).toBe(_.schemaComposer.getAnyTC('Foo'));
      });
      it('should accept type as string to union type', () => {
        _ObjectTypeComposer.ObjectTypeComposer.create('type A { id: ID! }', _.schemaComposer);

        _ObjectTypeComposer.ObjectTypeComposer.create('type B { id: ID! }', _.schemaComposer);

        _UnionTypeComposer.UnionTypeComposer.create('union Foo = A | B', _.schemaComposer);

        const fc = typeMapper.convertOutputFieldConfig({
          type: 'Foo'
        });
        expect(fc.type).toBe(_.schemaComposer.getAnyTC('Foo'));
      });
      it('should create field config from type as string', () => {
        const fc = typeMapper.convertOutputFieldConfig('String');
        expect(fc.type).toBeInstanceOf(_ScalarTypeComposer.ScalarTypeComposer);
        expect(fc.type.getType()).toBe(_graphql.GraphQLString);
      });
      it('should lookup type name as string in schemaComposer', () => {
        const tc = _ObjectTypeComposer.ObjectTypeComposer.create(`type MyType { a: Int }`, _.schemaComposer);

        const fc = typeMapper.convertOutputFieldConfig('MyType');
        expect(fc.type).toBe(tc);
        const fc2 = typeMapper.convertOutputFieldConfig({
          type: '[MyType]'
        });
        expect(fc2.type).toBeInstanceOf(_ListComposer.ListComposer);
        expect(fc2.type.getType().ofType).toBe(tc.getType());
      });
      it('should create field config from GraphQL Schema Language', () => {
        const fc = typeMapper.convertOutputFieldConfig(`type MyOutputType {
          a: String,
          b: Int,
        }`);
        const tc = fc.type;
        expect(tc.getTypeName()).toBe('MyOutputType');
        expect(tc.getFieldType('a')).toBe(_graphql.GraphQLString);
        expect(tc.getFieldType('b')).toBe(_graphql.GraphQLInt);
      });
      it('should create field with Scalar type from GraphQL Schema Language', () => {
        const fc = typeMapper.convertOutputFieldConfig('scalar MyScalar');
        expect(fc.type).toBeInstanceOf(_ScalarTypeComposer.ScalarTypeComposer);
        expect(fc.type.getTypeName()).toBe('MyScalar');
      });
      it('should create field with Enum type from GraphQL Schema Language', () => {
        const fc = typeMapper.convertOutputFieldConfig('enum MyEnum { AND OR }');
        expect(fc.type).toBeInstanceOf(_EnumTypeComposer.EnumTypeComposer);
        const enumValues = fc.type.getFields();
        expect(enumValues.AND).toMatchObject({
          value: 'AND'
        });
        expect(enumValues.OR).toMatchObject({
          value: 'OR'
        });
      });
      it('should throw error if provided input type definition', () => {
        expect(() => {
          typeMapper.convertOutputFieldConfig(`input MyInputType {
            a: String,
          }`);
        }).toThrowError(/but got input type definition/);
      });
      it('should accept ObjectTypeComposer', () => {
        const tc = _ObjectTypeComposer.ObjectTypeComposer.create('type PriceRange { lon: Float, lat: Float }', _.schemaComposer);

        tc.setDescription('Description');
        const fc = typeMapper.convertOutputFieldConfig({
          type: tc
        });
        expect(fc.type).toBe(tc);
        expect(fc.description).toBe(undefined);
        const fc2 = typeMapper.convertOutputFieldConfig(tc);
        expect(fc2.type).toBe(tc);
        expect(fc2.description).toBe(undefined);
      });
      it('should accept ScalarTypeComposer', () => {
        const stc = _ScalarTypeComposer.ScalarTypeComposer.create('scalar UInt', _.schemaComposer);

        stc.setDescription('Description');
        const fc = typeMapper.convertOutputFieldConfig({
          type: stc
        });
        expect(fc.type).toBe(stc);
        expect(fc.description).toBe(undefined);
        const fc2 = typeMapper.convertOutputFieldConfig(stc);
        expect(fc2.type).toBe(stc);
        expect(fc2.description).toBe(undefined);
      });
      it('should accept EnumTypeComposer', () => {
        const etc = _EnumTypeComposer.EnumTypeComposer.create('enum MyEnum { V1 V2 V3 }', _.schemaComposer);

        etc.setDescription('Description');
        const fc = typeMapper.convertOutputFieldConfig({
          type: etc
        });
        expect(fc.type).toBe(etc);
        expect(fc.description).toBe(undefined);
        const fc2 = typeMapper.convertOutputFieldConfig(etc);
        expect(fc2.type).toBe(etc);
        expect(fc2.description).toBe(undefined);
      });
      it('should accept InterfaceTypeComposer', () => {
        const iftc = _InterfaceTypeComposer.InterfaceTypeComposer.create('interface MyIFace { id: Int }', _.schemaComposer);

        iftc.setDescription('Description');
        const fc = typeMapper.convertOutputFieldConfig({
          type: iftc
        });
        expect(fc.type).toBe(iftc);
        expect(fc.description).toBe(undefined);
        const fc2 = typeMapper.convertOutputFieldConfig(iftc);
        expect(fc2.type).toBe(iftc);
        expect(fc2.description).toBe(undefined);
      });
      it('should accept Resolver', () => {
        const resolver = new _Resolver.Resolver({
          name: 'find',
          type: 'Float',
          args: {
            a1: 'String'
          },
          resolve: () => 123
        }, _.schemaComposer);
        const fc = typeMapper.convertOutputFieldConfig(resolver);
        expect(fc.type).toBeInstanceOf(_ScalarTypeComposer.ScalarTypeComposer);
        expect(fc.type.getType()).toBe(_graphql.GraphQLFloat);
        expect(fc.args.a1.type.getType()).toBe(_graphql.GraphQLString);
        expect(fc.resolve()).toBe(123);
      });
      it('should accept Resolver as type', () => {
        const resolver = new _Resolver.Resolver({
          name: 'find',
          type: 'Float',
          args: {
            a1: 'String'
          },
          resolve: () => 123
        }, _.schemaComposer);
        const fc = typeMapper.convertOutputFieldConfig({
          type: resolver
        });
        expect(fc.type).toBeInstanceOf(_ScalarTypeComposer.ScalarTypeComposer);
        expect(fc.type.getType()).toBe(_graphql.GraphQLFloat);
        expect(fc.args).toEqual({});
        expect(fc.resolve).toBe(undefined);
      });
      it('should pass unchanged thunk', () => {
        const myTypeThunk = () => 'Int';

        const fc = typeMapper.convertOutputFieldConfig(myTypeThunk);
        expect(fc.type).toBeInstanceOf(_ScalarTypeComposer.ScalarTypeComposer);
        expect(fc.type.getType()).toBe(_graphql.GraphQLInt);
      });
      it('should accept array with one element as type and wrap them with GraphQLList', () => {
        const fc = typeMapper.convertOutputFieldConfig(['String']);
        expect(fc.type).toBeInstanceOf(_ListComposer.ListComposer);
        expect(fc.type.ofType).toBeInstanceOf(_ScalarTypeComposer.ScalarTypeComposer);
        expect(fc.type.ofType.getType()).toBe(_graphql.GraphQLString);
        const fc2 = typeMapper.convertOutputFieldConfig({
          type: ['String']
        });
        expect(fc2.type).toBeInstanceOf(_ListComposer.ListComposer);
        expect(fc2.type.getType().ofType).toBe(_graphql.GraphQLString);
        const fc3 = typeMapper.convertOutputFieldConfig({
          type: [_graphql.GraphQLString]
        });
        expect(fc3.type).toBeInstanceOf(_ListComposer.ListComposer);
        expect(fc3.type.getType().ofType).toBe(_graphql.GraphQLString);

        const tc = _ObjectTypeComposer.ObjectTypeComposer.create('type PriceRange { lon: Float, lat: Float }', _.schemaComposer);

        const fc4 = typeMapper.convertOutputFieldConfig([tc]);
        expect(fc4.type).toBeInstanceOf(_ListComposer.ListComposer);
        expect(fc4.type.getType().ofType).toBe(tc.getType());
        const fc5 = typeMapper.convertOutputFieldConfig({
          type: [tc]
        });
        expect(fc5.type).toBeInstanceOf(_ListComposer.ListComposer);
        expect(fc5.type.ofType).toBe(tc);
        expect(() => {
          typeMapper.convertOutputFieldConfig([]);
        }).toThrowError(/Array must have exact one output type definition/);
        const fc6 = typeMapper.convertOutputFieldConfig([['String']]);
        expect(fc6.type).toBeInstanceOf(_ListComposer.ListComposer);
        expect(fc6.type.getType().ofType).toBeInstanceOf(_graphql.GraphQLList);
        expect(fc6.type.getType().ofType.ofType).toBe(_graphql.GraphQLString);
      });
      it('should throw error if provided InputTypeComposer', () => {
        const itc = _InputTypeComposer.InputTypeComposer.create('input LonLatInput { lon: Float, lat: Float }', _.schemaComposer);

        expect(() => {
          typeMapper.convertOutputFieldConfig({
            type: itc
          });
        }).toThrowError(/InputTypeComposer/);
        expect(() => {
          typeMapper.convertOutputFieldConfig(itc);
        }).toThrowError(/InputTypeComposer/);
      });
    });
    it('should convert args types', () => {
      const fc = typeMapper.convertOutputFieldConfig({
        type: 'String',
        args: {
          a1: {
            type: 'String'
          },
          a2: 'Int'
        }
      });
      expect(fc.args.a1.type).toBeInstanceOf(_ScalarTypeComposer.ScalarTypeComposer);
      expect(fc.args.a1.type.getType()).toBe(_graphql.GraphQLString);
      expect(fc.args.a2.type.getType()).toBe(_graphql.GraphQLInt);
    });
    it('should process outputFieldConfigMap()', () => {
      const fcm = typeMapper.convertOutputFieldConfigMap({
        f1: 'String',
        f2: 'Int'
      });
      expect(fcm.f1.type).toBeInstanceOf(_ScalarTypeComposer.ScalarTypeComposer);
      expect(fcm.f1.type.getType()).toBe(_graphql.GraphQLString);
      expect(fcm.f2.type.getType()).toBe(_graphql.GraphQLInt);
    });
  });
  describe('convertInputFieldConfig()', () => {
    it('should accept type with GraphQLInputObjectType', () => {
      const ic = typeMapper.convertInputFieldConfig({
        type: _graphql.GraphQLString
      });
      expect(ic.type).toBeInstanceOf(_ScalarTypeComposer.ScalarTypeComposer);
      expect(ic.type.getType()).toBe(_graphql.GraphQLString);
      const objectType = new _graphql.GraphQLInputObjectType({
        name: 'SomeTypeInput',
        fields: {
          f: {
            type: _graphql.GraphQLString
          }
        }
      });
      const ic2 = typeMapper.convertInputFieldConfig({
        type: objectType
      });
      expect(ic2.type).toBeInstanceOf(_InputTypeComposer.InputTypeComposer);
      expect(ic2.type.getType()).toBe(objectType);
    });
    it('should accept GraphQLScalarType', () => {
      const ic = typeMapper.convertInputFieldConfig(_graphql.GraphQLString);
      expect(ic.type.getType()).toBe(_graphql.GraphQLString);
    });
    it('should accept GraphQLInputObjectType', () => {
      const type = new _graphql.GraphQLInputObjectType({
        name: 'InputType',
        fields: () => ({
          f1: {
            type: _graphql.GraphQLInt
          }
        })
      });
      const ic = typeMapper.convertInputFieldConfig(type);
      expect(ic.type).toBeInstanceOf(_InputTypeComposer.InputTypeComposer);
      expect(ic.type.getType()).toBe(type);
    });
    it('should accept GraphQLNonNull', () => {
      const ic = typeMapper.convertInputFieldConfig(new _graphql.GraphQLNonNull(_graphql.GraphQLString));
      expect(ic.type).toBeInstanceOf(_NonNullComposer.NonNullComposer);
      expect(ic.type.getType().ofType).toBe(_graphql.GraphQLString);
    });
    it('should accept GraphQLList', () => {
      const ic = typeMapper.convertInputFieldConfig(new _graphql.GraphQLList(_graphql.GraphQLString));
      expect(ic.type).toBeInstanceOf(_ListComposer.ListComposer);
      expect(ic.type.getType().ofType).toBe(_graphql.GraphQLString);
    });
    it('should accept type as string to Scalar', () => {
      const ic = typeMapper.convertInputFieldConfig({
        type: 'String'
      });
      expect(ic.type.getType()).toBe(_graphql.GraphQLString);
    });
    it('should accept type as string to Enum', () => {
      _EnumTypeComposer.EnumTypeComposer.create('enum Foo { A B }', _.schemaComposer);

      const ic = typeMapper.convertInputFieldConfig({
        type: 'Foo'
      });
      expect(ic.type).toBe(_.schemaComposer.get('Foo'));
    });
    it('should accept type as string to Input Object', () => {
      _InputTypeComposer.InputTypeComposer.create('input Foo { id: Sting }', _.schemaComposer);

      const ic = typeMapper.convertInputFieldConfig({
        type: 'Foo'
      });
      expect(ic.type).toBe(_.schemaComposer.get('Foo'));
    });
    it('should create field config from type name as string', () => {
      const ic = typeMapper.convertInputFieldConfig('String');
      expect(ic.type).toBeInstanceOf(_ScalarTypeComposer.ScalarTypeComposer);
      expect(ic.type.getType()).toBe(_graphql.GraphQLString);
    });
    it('should lookup type name as string in schemaComposer', () => {
      const itc = _InputTypeComposer.InputTypeComposer.create(`input MyInput { a: Int }`, _.schemaComposer);

      const ic = typeMapper.convertInputFieldConfig('MyInput');
      expect(ic.type).toBe(itc);
    });
    it('should create field config from input type GraphQL Schema Language', () => {
      const fc = typeMapper.convertInputFieldConfig(`input MyInputType {
          a: String,
          b: Int,
        }`);
      const itc = fc.type;
      expect(itc.getTypeName()).toBe('MyInputType');
      expect(itc.getFieldType('a')).toBe(_graphql.GraphQLString);
      expect(itc.getFieldType('b')).toBe(_graphql.GraphQLInt);
    });
    it('should create field with Scalar type from GraphQL Schema Language', () => {
      const fc = typeMapper.convertInputFieldConfig('scalar MyInput');
      expect(fc.type).toBeInstanceOf(_ScalarTypeComposer.ScalarTypeComposer);
      expect(fc.type.getTypeName()).toBe('MyInput');
    });
    it('should create field with Enum type from GraphQL Schema Language', () => {
      const fc = typeMapper.convertInputFieldConfig('enum MyInputEnum { AND OR }');
      expect(fc.type).toBeInstanceOf(_EnumTypeComposer.EnumTypeComposer);
      const enumValues = fc.type.getFields();
      expect(enumValues.AND).toMatchObject({
        value: 'AND'
      });
      expect(enumValues.OR).toMatchObject({
        value: 'OR'
      });
    });
    it('should throw error if provided output type definition', () => {
      expect(() => {
        typeMapper.convertInputFieldConfig(`type MyOutputType {
          a: String,
        }`);
      }).toThrowError(/but got output type definitio/);
    });
    it('should accept InputTypeComposer', () => {
      const itc = _InputTypeComposer.InputTypeComposer.create('input PriceRangeInput { lon: Float, lat: Float }', _.schemaComposer);

      itc.setDescription('Description');
      const ic = typeMapper.convertInputFieldConfig({
        type: itc
      });
      expect(ic.type).toBe(itc);
      expect(ic.description).toBe(undefined);
      const ic2 = typeMapper.convertInputFieldConfig(itc);
      expect(ic2.type).toBe(itc);
      expect(ic2.description).toBe(undefined);
    });
    it('should accept ScalarTypeComposer', () => {
      const stc = _.schemaComposer.createScalarTC('scalar MySca');

      stc.setDescription('Description');
      const ic = typeMapper.convertInputFieldConfig({
        type: stc
      });
      expect(ic.type).toBe(stc);
      expect(ic.description).toBe(undefined);
      const ic2 = typeMapper.convertInputFieldConfig(stc);
      expect(ic2.type).toBe(stc);
      expect(ic2.description).toBe(undefined);
    });
    it('should accept EnumTypeComposer', () => {
      const etc = _.schemaComposer.createEnumTC('enum MyEnum { V1 V2 }');

      etc.setDescription('Description');
      const ic = typeMapper.convertInputFieldConfig({
        type: etc
      });
      expect(ic.type).toBe(etc);
      expect(ic.description).toBe(undefined);
      const ic2 = typeMapper.convertInputFieldConfig(etc);
      expect(ic2.type).toBe(etc);
      expect(ic2.description).toBe(undefined);
    });
    it('should throw error if provided ObjectTypeComposer', () => {
      const tc = _.schemaComposer.createObjectTC('type LonLat { lon: Float, lat: Float }');

      expect(() => {
        typeMapper.convertInputFieldConfig({
          type: tc
        });
      }).toThrowError(/\sObjectTypeComposer/);
      expect(() => {
        typeMapper.convertInputFieldConfig(tc);
      }).toThrowError(/\sObjectTypeComposer/);
    });
    it('should pass unchanged thunk', () => {
      const myTypeThunk = {
        type: () => 'Int'
      };
      const tc = typeMapper.convertInputFieldConfig(myTypeThunk);
      expect(tc.type).toBeInstanceOf(_ThunkComposer.ThunkComposer);
      expect(tc.type._thunk()).toBeInstanceOf(_ScalarTypeComposer.ScalarTypeComposer);
      expect(tc.type._thunk().getType()).toBe(_graphql.GraphQLInt);
    });
    it('should accept array with one element as type and wrap them with GraphQLList', () => {
      const fc = typeMapper.convertInputFieldConfig(['String']);
      expect(fc.type).toBeInstanceOf(_ListComposer.ListComposer);
      expect(fc.type.getType().ofType).toBe(_graphql.GraphQLString);
      const fc2 = typeMapper.convertInputFieldConfig({
        type: ['String']
      });
      expect(fc2.type).toBeInstanceOf(_ListComposer.ListComposer);
      expect(fc2.type.getType().ofType).toBe(_graphql.GraphQLString);
      const fc3 = typeMapper.convertInputFieldConfig({
        type: [_graphql.GraphQLString]
      });
      expect(fc3.type).toBeInstanceOf(_ListComposer.ListComposer);
      expect(fc3.type.getType().ofType).toBe(_graphql.GraphQLString);

      const itc = _InputTypeComposer.InputTypeComposer.create('input PriceRangeInput { lon: Float, lat: Float }', _.schemaComposer);

      const fc4 = typeMapper.convertInputFieldConfig([itc]);
      expect(fc4.type).toBeInstanceOf(_ListComposer.ListComposer);
      expect(fc4.type.getType().ofType).toBe(itc.getType());
      const fc5 = typeMapper.convertInputFieldConfig({
        type: [itc]
      });
      expect(fc5.type).toBeInstanceOf(_ListComposer.ListComposer);
      expect(fc5.type.getType().ofType).toBe(itc.getType());
      const fc6 = typeMapper.convertInputFieldConfig([['String']]);
      expect(fc6.type).toBeInstanceOf(_ListComposer.ListComposer);
      expect(fc6.type.ofType).toBeInstanceOf(_ListComposer.ListComposer);
      expect(fc6.type.ofType.ofType).toBeInstanceOf(_ScalarTypeComposer.ScalarTypeComposer);
      expect(fc6.type.getType().ofType.ofType).toBe(_graphql.GraphQLString);
      expect(() => {
        typeMapper.convertInputFieldConfig([]);
      }).toThrowError(/Array must have exact one input type definition/);
    });
    it('should process inputFieldConfigMap()', () => {
      const icm = typeMapper.convertInputFieldConfigMap({
        i1: {
          type: 'String'
        },
        i2: 'Int'
      });
      expect(icm.i1.type.getType()).toBe(_graphql.GraphQLString);
      expect(icm.i2.type.getType()).toBe(_graphql.GraphQLInt);
    });
  });
  describe('convertArgConfig()', () => {
    it('should accept type with GraphQLInputObjectType', () => {
      const ac = typeMapper.convertArgConfig({
        type: _graphql.GraphQLString
      });
      expect(ac.type).toBeInstanceOf(_ScalarTypeComposer.ScalarTypeComposer);
      expect(ac.type.getType()).toBe(_graphql.GraphQLString);
      const objectType = new _graphql.GraphQLInputObjectType({
        name: 'SomeTypeInput',
        fields: {
          f: {
            type: _graphql.GraphQLString
          }
        }
      });
      const ac2 = typeMapper.convertArgConfig({
        type: objectType
      });
      expect(ac2.type.getType()).toBe(objectType);
    });
    it('should accept GraphQLScalarType', () => {
      const ac = typeMapper.convertArgConfig(_graphql.GraphQLString);
      expect(ac.type).toBeInstanceOf(_ScalarTypeComposer.ScalarTypeComposer);
      expect(ac.type.getType()).toBe(_graphql.GraphQLString);
    });
    it('should accept GraphQLInputObjectType', () => {
      const type = new _graphql.GraphQLInputObjectType({
        name: 'InputType',
        fields: () => ({
          f: {
            type: _graphql.GraphQLInt
          }
        })
      });
      const ac = typeMapper.convertArgConfig(type);
      expect(ac.type).toBeInstanceOf(_InputTypeComposer.InputTypeComposer);
      expect(ac.type.getType()).toBe(type);
    });
    it('should accept GraphQLNonNull', () => {
      const ac = typeMapper.convertArgConfig(new _graphql.GraphQLNonNull(_graphql.GraphQLString));
      expect(ac.type).toBeInstanceOf(_NonNullComposer.NonNullComposer);
      expect(ac.type.getType().ofType).toBe(_graphql.GraphQLString);
    });
    it('should accept GraphQLList', () => {
      const ac = typeMapper.convertArgConfig(new _graphql.GraphQLList(_graphql.GraphQLString));
      expect(ac.type).toBeInstanceOf(_ListComposer.ListComposer);
      expect(ac.type.getType().ofType).toBe(_graphql.GraphQLString);
    });
    it('should accept type as string to Scalar', () => {
      const ac = typeMapper.convertArgConfig({
        type: 'String'
      });
      expect(ac.type.getType()).toBe(_graphql.GraphQLString);
    });
    it('should accept type as string to Enum', () => {
      _EnumTypeComposer.EnumTypeComposer.create('enum Foo { A B }', _.schemaComposer);

      const ac = typeMapper.convertArgConfig({
        type: 'Foo'
      });
      expect(ac.type).toBe(_.schemaComposer.get('Foo'));
    });
    it('should accept type as string to Input Object', () => {
      _InputTypeComposer.InputTypeComposer.create('input Foo { id: String }', _.schemaComposer);

      const ac = typeMapper.convertArgConfig({
        type: 'Foo'
      });
      expect(ac.type).toBe(_.schemaComposer.get('Foo'));
    });
    it('should create arg config from GraphQL Schema Language', () => {
      const ac = typeMapper.convertArgConfig('String');
      expect(ac.type.getType()).toBe(_graphql.GraphQLString);
    });
    it('should lookup type name as string in schemaComposer', () => {
      const itc = _InputTypeComposer.InputTypeComposer.create(`input MyArg { a: Int }`, _.schemaComposer);

      const ac = typeMapper.convertArgConfig('MyArg');
      expect(ac.type).toBe(itc);
    });
    it('should create arg config from input type GraphQL Schema Language', () => {
      const fc = typeMapper.convertArgConfig(`input MyInputArg {
        a: String,
        b: Int,
      }`);
      const itc = fc.type;
      expect(itc.getTypeName()).toBe('MyInputArg');
      expect(itc.getFieldType('a')).toBe(_graphql.GraphQLString);
      expect(itc.getFieldType('b')).toBe(_graphql.GraphQLInt);
    });
    it('should create arg config with Scalar type from GraphQL Schema Language', () => {
      const fc = typeMapper.convertArgConfig('scalar Abc');
      expect(fc.type).toBeInstanceOf(_ScalarTypeComposer.ScalarTypeComposer);
      expect(fc.type.getType()).toBeInstanceOf(_graphql.GraphQLScalarType);
      expect(fc.type.getTypeName()).toBe('Abc');
    });
    it('should create arg config with Enum type from GraphQL Schema Language', () => {
      const fc = typeMapper.convertArgConfig('enum MyArgEnum { AND OR }');
      expect(fc.type).toBeInstanceOf(_EnumTypeComposer.EnumTypeComposer);
      expect(fc.type.getType()).toBeInstanceOf(_graphql.GraphQLEnumType);
      const enumValues = fc.type.getFields();
      expect(enumValues.AND).toMatchObject({
        value: 'AND'
      });
      expect(enumValues.OR).toMatchObject({
        value: 'OR'
      });
    });
    it('should throw error if provided output type definition', () => {
      expect(() => {
        typeMapper.convertArgConfig(`type MyOutputType {
          a: String,
        }`);
      }).toThrowError(/but got output type definition/);
    });
    it('should accept InputTypeComposer', () => {
      const itc = _InputTypeComposer.InputTypeComposer.create('input PriceRangeInput { lon: Float, lat: Float }', _.schemaComposer);

      itc.setDescription('Description');
      const ac = typeMapper.convertArgConfig({
        type: itc
      });
      expect(ac.type).toBeInstanceOf(_InputTypeComposer.InputTypeComposer);
      expect(ac.type).toBe(itc);
      expect(ac.type.getType()).toBe(itc.getType());
      expect(ac.description).toBe(undefined);
      const ac2 = typeMapper.convertArgConfig(itc);
      expect(ac2.type.getType()).toBe(itc.getType());
      expect(ac2.description).toBe(undefined);
    });
    it('should accept ScalarTypeComposer', () => {
      const stc = _ScalarTypeComposer.ScalarTypeComposer.create('scalar Aaa', _.schemaComposer);

      stc.setDescription('Description');
      const ac = typeMapper.convertArgConfig({
        type: stc
      });
      expect(ac.type).toBeInstanceOf(_ScalarTypeComposer.ScalarTypeComposer);
      expect(ac.type).toBe(stc);
      expect(ac.type.getType()).toBe(stc.getType());
      expect(ac.description).toBe(undefined);
      const ac2 = typeMapper.convertArgConfig(stc);
      expect(ac2.type.getType()).toBe(stc.getType());
      expect(ac2.description).toBe(undefined);
    });
    it('should accept EnumTypeComposer', () => {
      const etc = _EnumTypeComposer.EnumTypeComposer.create('enum MyEnum { V1 V2 }', _.schemaComposer);

      etc.setDescription('Description');
      const ac = typeMapper.convertArgConfig({
        type: etc
      });
      expect(ac.type).toBeInstanceOf(_EnumTypeComposer.EnumTypeComposer);
      expect(ac.type).toBe(etc);
      expect(ac.type.getType()).toBe(etc.getType());
      expect(ac.description).toBe(undefined);
      const ac2 = typeMapper.convertArgConfig(etc);
      expect(ac2.type).toBeInstanceOf(_EnumTypeComposer.EnumTypeComposer);
      expect(ac2.type).toBe(etc);
      expect(ac2.type.getType()).toBe(etc.getType());
      expect(ac2.description).toBe(undefined);
    });
    it('should pass unchanged thunk', () => {
      const myTypeThunk = () => 'Int';

      const ac = typeMapper.convertArgConfig(myTypeThunk);
      expect(ac.type).toBeInstanceOf(_ScalarTypeComposer.ScalarTypeComposer);
      expect(ac.type.getType()).toBe(_graphql.GraphQLInt);
    });
    it('should accept array with one element as type and wrap them with GraphQLList', () => {
      const fc = typeMapper.convertArgConfig(['String']);
      expect(fc.type).toBeInstanceOf(_ListComposer.ListComposer);
      expect(fc.type.getType()).toBeInstanceOf(_graphql.GraphQLList);
      expect(fc.type.getType().ofType).toBe(_graphql.GraphQLString);
      const fc2 = typeMapper.convertArgConfig({
        type: ['String']
      });
      expect(fc2.type).toBeInstanceOf(_ListComposer.ListComposer);
      expect(fc2.type.getType()).toBeInstanceOf(_graphql.GraphQLList);
      expect(fc2.type.getType().ofType).toBe(_graphql.GraphQLString);
      const fc3 = typeMapper.convertArgConfig({
        type: [_graphql.GraphQLString]
      });
      expect(fc3.type).toBeInstanceOf(_ListComposer.ListComposer);
      expect(fc3.type.getType()).toBeInstanceOf(_graphql.GraphQLList);
      expect(fc3.type.getType().ofType).toBe(_graphql.GraphQLString);

      const itc = _InputTypeComposer.InputTypeComposer.create('input PriceRangeInput { lon: Float, lat: Float }', _.schemaComposer);

      const fc4 = typeMapper.convertArgConfig([itc]);
      expect(fc4.type).toBeInstanceOf(_ListComposer.ListComposer);
      expect(fc4.type.getType()).toBeInstanceOf(_graphql.GraphQLList);
      expect(fc4.type.getType().ofType).toBe(itc.getType());
      const fc5 = typeMapper.convertArgConfig({
        type: [itc]
      });
      expect(fc5.type).toBeInstanceOf(_ListComposer.ListComposer);
      expect(fc5.type.getType()).toBeInstanceOf(_graphql.GraphQLList);
      expect(fc5.type.getType().ofType).toBe(itc.getType());
      const fc6 = typeMapper.convertArgConfig([['String']]);
      expect(fc6.type).toBeInstanceOf(_ListComposer.ListComposer);
      expect(fc6.type.getType()).toBeInstanceOf(_graphql.GraphQLList);
      expect(fc6.type.getType().ofType).toBeInstanceOf(_graphql.GraphQLList);
      expect(fc6.type.getType().ofType.ofType).toBe(_graphql.GraphQLString);
      expect(() => {
        typeMapper.convertArgConfig([]);
      }).toThrowError(/Array must have exact one input type definition/);
    });
    it('should throw error if provided ObjectTypeComposer', () => {
      const tc = _ObjectTypeComposer.ObjectTypeComposer.create('type LonLat { lon: Float, lat: Float }', _.schemaComposer);

      expect(() => {
        typeMapper.convertArgConfig({
          type: tc
        });
      }).toThrowError(/\sObjectTypeComposer/);
      expect(() => {
        typeMapper.convertArgConfig(tc);
      }).toThrowError(/\sObjectTypeComposer/);
    });
    it('should process ArgConfigMap', () => {
      const acm = typeMapper.convertArgConfigMap({
        a1: {
          type: 'String'
        },
        a2: 'Int'
      });
      expect(acm.a1.type).toBeInstanceOf(_ScalarTypeComposer.ScalarTypeComposer);
      expect(acm.a1.type.getType()).toBe(_graphql.GraphQLString);
      expect(acm.a2.type).toBeInstanceOf(_ScalarTypeComposer.ScalarTypeComposer);
      expect(acm.a2.type.getType()).toBe(_graphql.GraphQLInt);
    });
  });
  describe('parseTypesFrom... methods', () => {
    it('parseTypesFromString()', () => {
      const gql = `
        type User {
          name: String
        }

        type Article {
          title: String
        }

        input Record {
          id: ID
          ts: Int
        }
      `;
      const ts = typeMapper.parseTypesFromString(gql);
      expect(Array.from(ts.keys())).toEqual(['User', 'Article', 'Record']);
      expect(ts.get('User').getType()).toBeInstanceOf(_graphql.GraphQLObjectType);
      expect(ts.get('Article').getType()).toBeInstanceOf(_graphql.GraphQLObjectType);
      expect(ts.get('Record').getType()).toBeInstanceOf(_graphql.GraphQLInputObjectType);
    });
    it('parseTypesFromString() should strictly accept `schema` definition', () => {
      const ts = typeMapper.parseTypesFromString(`
        schema {
          query: Query
          mutation: Mutation
          subscription: Subscription
        }
      `);
      expect(ts.size).toEqual(0);
      expect(() => {
        typeMapper.parseTypesFromString(`schema { query: ErrName }`);
      }).toThrow("Incorrect type name 'ErrName' for 'query'");
      expect(() => {
        typeMapper.parseTypesFromString(`schema { mutation: ErrName }`);
      }).toThrow("Incorrect type name 'ErrName' for 'mutation'");
      expect(() => {
        typeMapper.parseTypesFromString(`schema { subscription: ErrName }`);
      }).toThrow("Incorrect type name 'ErrName' for 'subscription'");
    });
    it('parseTypesFromString() should strictly accept `scalar` definition', () => {
      const ts = typeMapper.parseTypesFromString(`scalar MyScalar`);
      expect(ts.get('MyScalar').getType()).toBeInstanceOf(_graphql.GraphQLScalarType);
    });
    it('parseTypesFromString() should accept `union` definition', () => {
      const ts = typeMapper.parseTypesFromString(`
        union TypeAB = TypeA | TypeB
        type TypeA { f1: Int }
        type TypeB { f2: Int }
      `);
      const TypeAB = ts.get('TypeAB');
      expect(TypeAB).toBeInstanceOf(_UnionTypeComposer.UnionTypeComposer);
      const types = TypeAB.getTypes();
      expect(types).toHaveLength(2);
    });
  });
  describe('convertSDLTypeDefinition()', () => {
    it('should return same type for the same TypeDefinitionString', () => {
      const t1 = typeMapper.convertSDLTypeDefinition('type SameType { a: Int }');
      const t2 = typeMapper.convertSDLTypeDefinition('type SameType { a: Int }');
      expect(t1).toBe(t2);
      expect(t1.getTypeName()).toBe('SameType');
      expect(t1.getFieldType('a')).toBe(_graphql.GraphQLInt);
    });
  });
  describe('parse types directives to extensions', () => {
    it('ObjectType', () => {
      const tc = typeMapper.convertSDLTypeDefinition(`
        type My1 @typeDirective(a: false) { 
          a: Int @default(value: 1)
          b: Float @unexisted(q: 1, w: true, e: "s")
          c: String @me @they @me
          d(arg: Int = 15 @darg(v: 2) @darg2(w: "3")): Int @ddd
        }`);
      expect(tc).toBeInstanceOf(_ObjectTypeComposer.ObjectTypeComposer);
      expect(tc.getExtensions()).toEqual({
        directives: [{
          args: {
            a: false
          },
          name: 'typeDirective'
        }]
      });
      expect(tc.getFieldExtensions('a')).toEqual({
        directives: [{
          args: {
            value: 1
          },
          name: 'default'
        }]
      });
      expect(tc.getFieldExtensions('b')).toEqual({
        directives: [{
          args: {
            e: 's',
            q: 1,
            w: true
          },
          name: 'unexisted'
        }]
      });
      expect(tc.getFieldExtensions('c')).toEqual({
        directives: [{
          args: {},
          name: 'me'
        }, {
          args: {},
          name: 'they'
        }, {
          args: {},
          name: 'me'
        }]
      });
      expect(tc.getFieldArgExtensions('d', 'arg')).toEqual({
        directives: [{
          args: {
            v: 2
          },
          name: 'darg'
        }, {
          args: {
            w: '3'
          },
          name: 'darg2'
        }]
      });
      expect(tc.getFieldExtensions('d')).toEqual({
        directives: [{
          args: {},
          name: 'ddd'
        }]
      });
    });
    it('InterfaceType', () => {
      const tc = typeMapper.convertSDLTypeDefinition(`
        interface My1 @typeDirective(a: false) { 
          a: Int @default(value: 1)
          b: Float @unexisted(q: 1, w: true, e: "s")
          c: String @me @they @me
          d(arg: Int = 15 @darg(v: 2) @darg2(w: "3")): Int @ddd
        }`);
      expect(tc).toBeInstanceOf(_InterfaceTypeComposer.InterfaceTypeComposer);
      expect(tc.getExtensions()).toEqual({
        directives: [{
          args: {
            a: false
          },
          name: 'typeDirective'
        }]
      });
      expect(tc.getFieldExtensions('a')).toEqual({
        directives: [{
          args: {
            value: 1
          },
          name: 'default'
        }]
      });
      expect(tc.getFieldExtensions('b')).toEqual({
        directives: [{
          args: {
            e: 's',
            q: 1,
            w: true
          },
          name: 'unexisted'
        }]
      });
      expect(tc.getFieldExtensions('c')).toEqual({
        directives: [{
          args: {},
          name: 'me'
        }, {
          args: {},
          name: 'they'
        }, {
          args: {},
          name: 'me'
        }]
      });
      expect(tc.getFieldArg('d', 'arg').defaultValue).toBe(15);
      expect(tc.getFieldArgExtensions('d', 'arg')).toEqual({
        directives: [{
          args: {
            v: 2
          },
          name: 'darg'
        }, {
          args: {
            w: '3'
          },
          name: 'darg2'
        }]
      });
      expect(tc.getFieldExtensions('d')).toEqual({
        directives: [{
          args: {},
          name: 'ddd'
        }]
      });
    });
    it('InputType', () => {
      const tc = typeMapper.convertSDLTypeDefinition(`
        input My1 @typeDirective(a: false) { 
          a: Int @default(value: 1)
          b: Float @unexisted(q: 1, w: true, e: "s")
          c: String @me @they @me
        }`);
      expect(tc).toBeInstanceOf(_InputTypeComposer.InputTypeComposer);
      expect(tc.getExtensions()).toEqual({
        directives: [{
          args: {
            a: false
          },
          name: 'typeDirective'
        }]
      });
      expect(tc.getFieldExtensions('a')).toEqual({
        directives: [{
          args: {
            value: 1
          },
          name: 'default'
        }]
      });
      expect(tc.getFieldExtensions('b')).toEqual({
        directives: [{
          args: {
            e: 's',
            q: 1,
            w: true
          },
          name: 'unexisted'
        }]
      });
      expect(tc.getFieldExtensions('c')).toEqual({
        directives: [{
          args: {},
          name: 'me'
        }, {
          args: {},
          name: 'they'
        }, {
          args: {},
          name: 'me'
        }]
      });
    });
  });
  describe('static methods', () => {
    it('check SDL types', () => {
      const output = 'type Out { name: String! }';
      const input = 'input Filter { minAge: Int }';
      const enumType = 'enum Sort { ASC DESC }';
      const scalar = 'scalar UInt';
      const iface = 'interface User { name: String }';
      expect(_TypeMapper.TypeMapper.isTypeDefinitionString(output)).toBeTruthy();
      expect(_TypeMapper.TypeMapper.isOutputTypeDefinitionString(output)).toBeTruthy();
      expect(_TypeMapper.TypeMapper.isOutputTypeDefinitionString(input)).toBeFalsy();
      expect(_TypeMapper.TypeMapper.isTypeDefinitionString(input)).toBeTruthy();
      expect(_TypeMapper.TypeMapper.isInputTypeDefinitionString(input)).toBeTruthy();
      expect(_TypeMapper.TypeMapper.isInputTypeDefinitionString(output)).toBeFalsy();
      expect(_TypeMapper.TypeMapper.isTypeDefinitionString(enumType)).toBeTruthy();
      expect(_TypeMapper.TypeMapper.isEnumTypeDefinitionString(enumType)).toBeTruthy();
      expect(_TypeMapper.TypeMapper.isEnumTypeDefinitionString(output)).toBeFalsy();
      expect(_TypeMapper.TypeMapper.isTypeDefinitionString(scalar)).toBeTruthy();
      expect(_TypeMapper.TypeMapper.isScalarTypeDefinitionString(scalar)).toBeTruthy();
      expect(_TypeMapper.TypeMapper.isScalarTypeDefinitionString(output)).toBeFalsy();
      expect(_TypeMapper.TypeMapper.isTypeDefinitionString(iface)).toBeTruthy();
      expect(_TypeMapper.TypeMapper.isInterfaceTypeDefinitionString(iface)).toBeTruthy();
      expect(_TypeMapper.TypeMapper.isInterfaceTypeDefinitionString(output)).toBeFalsy();
    });
    it('check type name', () => {
      expect(_TypeMapper.TypeMapper.isTypeNameString('aaaa')).toBeTruthy();
      expect(_TypeMapper.TypeMapper.isTypeNameString('Aaaaa')).toBeTruthy();
      expect(_TypeMapper.TypeMapper.isTypeNameString('A_')).toBeTruthy();
      expect(_TypeMapper.TypeMapper.isTypeNameString('_A')).toBeTruthy();
      expect(_TypeMapper.TypeMapper.isTypeNameString('A_123')).toBeTruthy();
      expect(_TypeMapper.TypeMapper.isTypeNameString('123')).toBeFalsy();
      expect(_TypeMapper.TypeMapper.isTypeNameString('A-')).toBeFalsy();
    });
  });
});